import random
from .lexigen import *
from .utils import *


class StochasticJolasticWordGenerator:
    common_words = ["the", "with", "in", "that", "not", "a", "an", "of", "for", "as", "like", "on", 'his', 'the',
                    'your', 'my', 'their']

    def __init__(self, previous_lines=[]):
        self.connector_choices = ['and', 'or', 'as', 'like', 'with']
        self.last_algorithms_used_to_reach_next_word = (None, None)
        self.previous_lines = previous_lines

    def random_nonrhyme(self, previous_words: List[str], rhymable: bool = False) -> str:
        """Return a random result of a random function that hits Project Datamuse API (rhyme function excluded)

        This function is primarily designed for use by the poem_line_from_markov function, but it may have other
        interesting uses. This is why it is designed to take all the previous words of a line, rather than just the
        last word. Some words do not have many/any result values for some of these API calls because they are
        too weird "weird" input values (i.e. proper nouns or slang), so to ensure the function returns a value,
        the function will try different random functions from the lexigen module with different words until a word is
        found that is both not too similar to preceding words and not too similar to the last line's last word.

        :param previous_words: an ordered list of previous words of generated poem line
        :param rhymable: result must have a valid rhyme
        """
        result = None
        while result is None:
            # Randomly choose up to two algorithms the next word - one's repeated in this list for increased probability
            next_word_algorithms = [similar_sounding_word, similar_meaning_word, contextually_linked_word,
                                    frequently_following_word, frequently_following_word]
            nw_algorithms_copy = next_word_algorithms.copy()
            if self.last_algorithms_used_to_reach_next_word and self.last_algorithms_used_to_reach_next_word[0]:
                # Don't use the same algorithm for picking two successive words
                next_word_algorithms.remove(self.last_algorithms_used_to_reach_next_word[0])
            random_algorithm = random.choice(next_word_algorithms)
            # The frequently following function should always use the preceding word as input
            # But otherwise, this will randomly sometimes use a different preceding word as input
            input_word = previous_words[-1] if random.random() <= .75 else random.choice(previous_words)
            if random_algorithm != frequently_following_word and random.random() <= .25:
                if self.last_algorithms_used_to_reach_next_word and self.last_algorithms_used_to_reach_next_word[1]:
                    # Same goe for the 2nd algorithm used though this should be pretty rare
                    nw_algorithms_copy.remove(self.last_algorithms_used_to_reach_next_word[1])
                second_random_algorithm = random.choice(nw_algorithms_copy)
                possible_result = random_algorithm(input_word)
                possible_result = second_random_algorithm(possible_result) if possible_result else \
                    second_random_algorithm(input_word)
                self.last_algorithms_used_to_reach_next_word = (random_algorithm, second_random_algorithm)
            else:
                possible_result = random_algorithm(input_word)
                self.last_algorithms_used_to_reach_next_word = (random_algorithm, None)
            if possible_result and not too_similar(possible_result, previous_words) and \
                    not (len(self.previous_lines) > 0 and
                         too_similar(possible_result, self.previous_lines[-1].split(' '))
                         and not has_invalid_characters(possible_result)) and \
                    not (rhymable and not rhyme(possible_result)):
                # Is the word too similar to another word in the line or the previous line?
                # Does the word have numbers or spaces for some reason? (extremely rare)
                # If so, keep trying; otherwise exit the loop and return the word)
                result = possible_result
        return result

    def last_word_of_markov_line(self, previous_words: List[str], rhyme_with: Optional[str] = None,
                                 max_length: Optional[int] = None) -> str:
        """Get the last word of a poem line generated by the markov algorithm and optionally try to make it rhyme.

        :param previous_words: an ordered list of previous words of generated poem line
        :param rhyme_with: the last word of the last line of the poem, if it exists
        :param max_line_legnth: an upper limit in characters for the word
        """
        word = None
        if rhyme_with:
            word = word = rhyme(rhyme_with)
            # if the word is a common word or would be awkward to end a line with, keep trying
            while word in self.common_words or (max_length and len(word) > max_length):
                word = rhyme(rhyme_with)
            # But if there's no rhyme result try another method altogether
            if not word:
                while word is None or (max_length and len(word) > max_length) or word in self.common_words \
                        or too_similar(word, previous_words):
                    word = self.random_nonrhyme(previous_words)
        else:
            while word is None or (max_length and len(word) > max_length) or word in self.common_words \
                    or too_similar(word, previous_words):
                # Maybe revisit defaulting rhymable to true here
                word = self.random_nonrhyme(previous_words, rhymable=True)
        return word

    def nonlast_word_of_markov_line(self, previous_words: List[str], words_for_sampling: List[str] = []) -> str:
        """Get the next word of a poem line generated by the markov algorithm.

        :param previous_words: an ordered list of previous words of generated poem line
        :param param words_for_sampling: a list of other words to throw in to the poem.
        """
        word = None
        if previous_words[-1] in self.common_words:
            if random.random() >= .85 and len(previous_words) > 1:
                word = self.random_nonrhyme(previous_words[:-1])
            else:
                while word is None or too_similar(word, previous_words):
                    word = random.choice(words_for_sampling)
        else:
            threshold = .6 if len(words_for_sampling) else 1
            while word is None or too_similar(word,  previous_words):
                if random.random() > threshold:
                    if random.random() <= .5:
                        word = random.choice(self.connector_choices)
                    else:
                        word = random.choice(words_for_sampling)
                else:
                    word = self.random_nonrhyme(previous_words)
        return word



import itertools
from typing import List, Optional
from .lexigen import *
from .jolastic import StochasticJolasticWordGenerator
from .utils import too_similar


class Poem:

    def __init__(self, input_words, words_for_sampling):
        self.input_words = input_words
        self.words_for_sampling = words_for_sampling
        self.title = "'".join(input_words)
        self.lines: List[str] = []

    def __str__(self):
        return self.raw_text

    def update(self):
        self.raw_text = '\n'.join(self.lines)

    @property
    def previous_line(self):
        if len(self.lines):
            return self.lines[-1]
        return ''


class PoemGenerator:

    def __init__(self):
        self.default_connectors = [' ', '   ', '...   ', random.choice([' & ', ' and ']), '  or  ', ' or ']
        self.line_enders = ['.', ', ', '!', '?', '', ' or', '...']
        self.markov_line_enders = ['', '', ',', ',', '!', '.', '?']
        self.line_indents = ['', '    ', '         ']
        self.poem = None


    def poem_line_from_markov(self, starting_word: str, num_words: int = 4, rhyme_with: Optional[str] = None,
                              words_for_sampling: List[str] = [], max_line_length: Optional[int] = 35) -> str:
        """Generate a line of poetry using a markov chain that optionally tries to make a line rhyme with the last one

        Different algorithms handle the last word and all the other words: both algorithms use a mix of random
        probability and process stopwords differently to keep the generated text interesting and non-repetitive.

        :param starting_word: the input word for the Markov algorithm, which hence is also the poem line's first word
        :param num_words: the number of words to write in the poem line
        :param rhyme_with: an optional word to try to make the poem line rhyme with. The algorithm will try something
                           else if this word is a common stopword or if it can't find a rhyme though.
        :param words_for_sampling: a list of other words to throw in to the poem. If you don't know what to pass here,
                                   phonetically related words to the starting word probably adds some sonority.
        :param max_line_length: an upper limit in characters for the line -- important for PDF generation to keep
                                everything on the page.
        """
        output_words, previous_word = [starting_word], starting_word
        markovgen = StochasticJolasticWordGenerator(previous_lines=self.poem.lines)
        for i in range(num_words - 1):
            if (i == num_words - 2) or (max_line_length and (max_line_length > 14 and
                                                             len(' '.join(output_words)) >= max_line_length - 14)):
                # Checks if if it's the last word--the limit can be determined by either word count or character count
                max_word_length = 12 if max_line_length else None
                word = markovgen.last_word_of_markov_line(output_words, rhyme_with=rhyme_with,
                                                          max_length=max_word_length)
                output_words.append(word)
                break
            else:
                word = markovgen.nonlast_word_of_markov_line(output_words, words_for_sampling=words_for_sampling)
                output_words.append(word)
        correct_a_vs_an(output_words)
        return " ".join(output_words)

    def poem_from_markov(self, input_words, num_lines=10, min_line_words: int = 5, max_line_words: int = 9,
                         max_line_length: Optional[int] = 35) -> str:
        """Generate a line of poetry using a markov chain that optionally tries to make a line rhyme with the last one
            Different algorithms handle the last word and all the other words: both algorithms use a mix of random
            probability and process stopwords differently to keep the generated text interesting and non-repetitive.

        :param input words: the user provided words to try making a poem from
        :param num_lines: the number of lines the poem will have
        :param max_line_words: the maximum number of words a line may have
        :param words_for_sampling: a list of other words to throw in to the poem. If you don't know what to pass here,
                                   phonetically related words to the starting word probably adds some sonority.
        :param max_line_length: an upper limit in characters for the line -- important for PDF generation to keep
                                everything on the page.
            """
        self.poem = None
        words_for_sampling = input_words + phonetically_related_words(input_words, limit_results_per_input_word=20)
        # Check for undesirable similarity overlap in the words for sampling list
        similarity_checks = list(itertools.combinations(words_for_sampling, 2))
        words_removed = []
        for word_pair in similarity_checks:
            if not(word_pair[0] in words_removed or word_pair[1] in words_removed) and \
                    too_similar(word_pair[0], word_pair[1]):
                words_removed.append(random.choice([word_pair[0], word_pair[1]]))
                words_for_sampling.remove(words_removed[-1])

        self.poem = Poem(input_words, words_for_sampling)
        last_line_last_word = ''
        random.shuffle(words_for_sampling)
        line_enders = []
        print("\n")
        for i in range(num_lines):
            rhyme_with = last_line_last_word if i % 2 == 1 else None
            # 67.5 % chance the line starts with an input word or something relate, 32.5% with a common word
            line_starter = words_for_sampling.pop() if random.random() > .4 else \
                    random.choice(StochasticJolasticWordGenerator.common_words)
            while i >= 1 and too_similar(line_starter, self.poem.lines[i - 1].split(' ')[0]):
                # while statement prevents repetition of line starters
                line_starter = words_for_sampling.pop() if random.random() > .4 else \
                    random.choice(StochasticJolasticWordGenerator.common_words)
            line = self.poem_line_from_markov(line_starter, words_for_sampling=words_for_sampling,
                                              num_words=random.randint(min_line_words, max_line_words),
                                              rhyme_with=rhyme_with, max_line_length=max_line_length)
            self.poem.lines.append(line)
            last_line_last_word = line.split(' ')[-1]
            # Directly adding line ender to line now will screw up rhyme pairs so save it & add it in another iteration
            line_enders.append(random.choice(self.markov_line_enders))
            print(line + line_enders[-1])
        for i, line in enumerate(self.poem.lines):
            self.poem.lines[i] += line_enders[i]
        poem = self.poem
        return poem

    def poem_line_from_word_list(self, word_list: List[str], max_line_length=35, connectors: List[str] = []) -> str:
        """Generate a line of a visual poem from a list of words by gluing them together with random connectors
           (whitespace, conjunctions, punctuation, and symbols).

        :param word_list: the words that will be used (in order, not randomly) that will form a visual poem
        :param max_line_length: upper limit on the length of the return value in characters
        :param connectors (list): list of glue strings
        """
        connectors = connectors if len(connectors) else self.default_connectors
        output, last_word = word_list[0], word_list[0]
        last_connector = ''
        for word in word_list[1:]:
            if random.random() < (
                    .2 + len(output) / 100):  # Increasing probability of line termination as line gets longer
                break
            if too_similar(last_word, word):
                continue
            connector = random.choice(connectors)
            while connector == last_connector:
                connector = random.choice(connectors)
            if len(output + connector + word) <= max_line_length:
                output += connector + word
            last_word = word
            last_connector = connector
        return output

    def poem_from_word_list(self, input_word_list: List[str], num_lines: int = 6, max_line_length: int = 35,
                            connectors: List[str] = [], limit_line_to_one_input_word: bool = False):
        """Generate a visual poem from a list of words by taking a given input word list, adding the phonetically
           related words to that word list, and then using those words to create a visual/concrete poem.

        :param input_word_list: the list of user-provided words that will be used, along with phonetically related
                                words, to generate a poem
        :param max_line_length: upper limit on the length of the return value in characters
        :param max_line_length: upper limit on length of poem lines (excluding line ending punctuation) in characters
        :param connectors: list of glue strings
        :param limit_line_to_one_input_word: If true, when generating a line of poetry, only use words that are
                                             phonetically related to one input word.
        """
        connectors = self.default_connectors if not len(connectors) else connectors
        output, line_indent = '', ''
        if limit_line_to_one_input_word:
            for i in range(num_lines - 1):
                linked_word = random.choice(input_word_list)
                output += self.poem_line_from_word_list(phonetically_related_words(linked_word), connectors=connectors,
                                                        max_line_length=max_line_length)
                line_indent = random.choice(self.line_indents) if line_indent == '' else \
                    random.choice([li for li in self.line_indents if li is not line_indent])  # Don't repeat the same indent 2x
                output += random.choice(self.line_enders) + '\n' + line_indent
        else:
            word_list = input_word_list.copy()
            for word in input_word_list:
                word_list.extend(phonetically_related_words(word))
            for i in range(num_lines - 1):
                random.shuffle(word_list)
                output += self.poem_line_from_word_list(word_list, connectors=connectors,
                                                        max_line_length=max_line_length)
                # Don't repeat the same indent 2x
                line_indent = random.choice(self.line_indents) if line_indent == '' else \
                    random.choice([li for li in self.line_indents if li is not line_indent])
                output += random.choice(self.line_enders) + '\n' + line_indent

        output += random.choice(input_word_list[:-1]) + ' ' + input_word_list[-1]
        return output


def print_poem(poem: str):
    """Print the poem with a newline before and after so it's easy to take a screenshot of its 'recipe' and the poem
    in your terminal and share it. :)

    :param poem: the poem, as a string, to be printed
    """
    print('\n')
    print(poem)
    print('\n')


from setuptools import setup, find_packages

__author__ = 'Corey Bobco'
__email__ = 'corey.bobco@gmail.com'
__version__ = '0.3.2'


with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = [
    'console-menu==0.6.0',
    'Gutenberg==0.8.0',
    'hunspell>=0.5.5',
    'inflect==4.1.0',
    'internetarchive==1.8.5',
    'markovify==0.8.0',
    'nltk==3.4.5',
    'pdf2image==1.12.1',
    'rdflib==4.2.2',
    'pronouncing>=0.2.0',
    'python-datamuse>=1.3.0',
    'spacy>=2.2.3',
    'reportlab>=3.5.26',
    'unittest2==1.1.0',
    'wordfreq>=2.2.2',
]

setup(
    name='generativepoetry',
    version='0.3.2',
    description='A library primarily for procedurally generating visual poems',
    long_description=readme,
    author="Corey Bobco",
    author_email='corey.bobco@gmail.com',
    url='https://github.com/coreybobco/generativepoetry-py',
    packages=[
        'generativepoetry',
    ],
    dependency_links = ['https://github.com/coreybobco/gutenberg_cleaner@master#egg=gutenberg_cleaner'],
    package_dir={'generativepoetry':
                 'generativepoetry'},
    package_data={'generativepoetry': ['data/hate_words.txt', 'data/abbreviations_etc.txt']},
    install_requires=requirements,
    scripts=['bin/generative-poetry-cli'],
    license="MIT",
    zip_safe=True,
    keywords='poetry',
    classifiers=[
        "Development Status :: 3 - Alpha",
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Topic :: Artistic Software",
    ],
    test_suite='tests',
)



import random
import re
from collections import defaultdict
from typing import List, TypeVar
import inflect
import markovify
import nltk
import spacy
from gutenberg.acquire import load_etext
from gutenberg.query import get_metadata
from gutenberg.cleanup import strip_headers
from gutenberg_cleaner import super_cleaner
from internetarchive import download
from urllib.parse import urlsplit


sent_detector = nltk.data.load('tokenizers/punkt/english.pickle')
spacy_nlp = spacy.load('en_core_web_sm', disable=['ner'])
spacy_nlp.remove_pipe("parser")
inflector = inflect.engine()
input_type = TypeVar('input_type', str, List[str])  # Must be str or list of strings


class ParsedText:

    def __init__(self, text):
        self.raw_text = text
        self.sentences = sent_detector.tokenize(text)
        self.paragraphs = self.raw_text.split("\n\n")

    def random_sentence(self, minimum_tokens=1) -> str:
        """Returns a random sentence from the text.

        Keyword Arguments:
            minimum_tokens; allows for sampling a sentence of a minimum NLP tokens
        """
        num_tokens = 0
        while num_tokens < minimum_tokens:
            sentence = random.choice(self.sentences)
            num_tokens = len([token.text for token in spacy_nlp(sentence)])
        return sentence

    def random_sentences(self, num=5, minimum_tokens=1) -> list:
        """Returns a random sentence from the text.

        Keyword Arguments:
            minimum_tokens; allows for sampling a sentence of a minimum NLP tokens
        """
        random_sentences = []
        while len(random_sentences) < num:
            random_sentence = self.random_sentence(minimum_tokens=minimum_tokens)
            if random_sentence not in random_sentences:
                random_sentences.append(random_sentence)
        return random_sentences

    def random_paragraph(self, minimum_sentences=3) -> str:
        """Returns a random sentence from the text.

        Keyword Arguments:
            minimum_tokens; allows for sampling a sentence of a minimum NLP tokens
        """
        num_sentences = 0
        while num_sentences < minimum_sentences:
            paragraph = random.choice(self.paragraphs)
            num_sentences = len(sent_detector.tokenize(paragraph))
        return paragraph


def validate_url(url, expected_netloc=''):
    """Validate that the provided string is indeed a URL from the anticipated source

     Keyword arguments:
        expected_netloc (str) -- the expected site the URL should be from, i.e. archive.org or gutenberg.org
    """
    url_parts = urlsplit(url)
    if not url_parts.netloc or (expected_netloc and expected_netloc not in url_parts.netloc):
        raise Exception(f'Not a valid f{expected_netloc} document url')


def get_internet_archive_document(url) -> str:
    """Downloads a document (book, etc.) from Internet Archive and returns it as a string. The linked document must
       have a text version. PDF text extraction is not supported at this time.
       Returns a ParsedText instance.
    """
    validate_url(url, expected_netloc='archive.org')
    url_parts = urlsplit(url).path.split("/")
    if len(url_parts) > 2:
        document_id = url_parts[2]
    else:
        raise Exception(f'Not a valid url')
    try:
        response = download(document_id, glob_pattern="*txt", return_responses=True)[0]
        # Remove single newlines, preserve double  newlines (because they demarcate paragraphs
        text = re.sub('(?<![\r\n])(\r?\n|\n?\r)(?![\r\n])', ' ', response.text.strip())
        # This usually creates double spaces between lines because most lines end with single spaces, but to account
        # for cases in which lines end without spaces, we will handle this in two lines
        return re.sub('(?<=[\S])(\s\s)(?=[\S])', ' ', text)

    except Exception:
        raise Exception(f'Archive.org download failed for url: {url}')


def get_gutenberg_document(url) -> str:
    """Downloads a document (book, etc.) from Project Gutenberg and returns it as a string.

    Returns a ParsedText instance."""
    # Get Project Gutenberg document ID from url string
    validate_url(url, expected_netloc='gutenberg.org')
    match = re.search("(?:files|ebooks|epub)\/(\d+)", urlsplit(url).path)
    if not match:
        raise Exception('Not a valid url')
    document_id = int(match.group(1))
    return super_cleaner(strip_headers(load_etext(document_id).strip()), mark_deletions=False)


def random_gutenberg_document(language_filter='en') -> str:
    """Downloads a random document (book, etc.) from Project Gutenberg and returns it as a string.

    Keyword arguments:
        language_filter (str) -- restrict the random document to a paritcular language (default: English)
    """
    doc_language = None
    document = ''
    while (not doc_language or language_filter) and doc_language != language_filter and len(document) == 0:
        # Keep grabbing random documents until 1 meets the language filter, if specified, and verify it really has text
        document_id = random.randint(1, 60134)  # Pick book at random (max id is currently 60134)
        lang_metadata = get_metadata('language', document_id)
        doc_language = next(iter(lang_metadata)) if len(lang_metadata) else False
        document = super_cleaner(strip_headers(load_etext(document_id).strip()), mark_deletions=False)
    return document


def reconcile_replacement_word(original_word_with_ws, original_word_tag, replacement_word, replacement_word_tag) -> str:
    """Modify replacement word if needed to fix subject/verb agreement and preserve the whitespace or lack of before
    and after the original word.

    Arguments:
        original_word_with_ws (str): (str) original word with surrounding whitespace
        original_word_tag (str): part-of-speech tag of original word
        replacement_word (str): word that is replacing original word
        replacement_word_tag (str):  part-of-speech tag of replacement word
    """
    # Pluralize or singularize the replacement word if we're dealing with nouns and one's plural and one's singular.
    if original_word_tag == 'NNS' and replacement_word_tag == 'NN':
        replacement_word = inflector.plural(replacement_word)
    elif original_word_tag == 'NN' and replacement_word_tag == 'NNS':
        replacement_word = inflector.singular_noun(replacement_word) \
            if inflector.singular_noun(replacement_word) else replacement_word
    #  Use regex to preserve the whitespace of the word-to-be-replaced
    replacement_word = re.sub('(?<!\S)\S+(?!\S)', replacement_word, original_word_with_ws)
    return replacement_word


def swap_parts_of_speech(text1, text2, parts_of_speech=['ADJ', 'NOUN']) -> (str, str):
    """Swap all the words of certain parts of speech from one text with those (with the same part of speech) from
    another text.

    Keyword arguments:
        parts_of_speech (list) -- list of parts of speech tags to swap out. Must be from the list provided by spaCy:
                                  https://spacy.io/api/annotation#pos-tagging
    """
    doc1 = spacy_nlp(text1)
    doc2 = spacy_nlp(text2)
    # First build two dictionaries (one for each text) whose keys are parts of speech and values are lists of words
    doc1_words_keyed_by_pos, doc2_words_keyed_by_pos = defaultdict(lambda: []), defaultdict(lambda: [])
    for token in doc1:
        if token.pos_ in parts_of_speech and not token.text in doc1_words_keyed_by_pos[token.pos_]:
            doc1_words_keyed_by_pos[token.pos_].append((token.text, token.tag_))
    for pos in parts_of_speech:
        random.shuffle(doc1_words_keyed_by_pos[pos])  # For variety's sake
    # Also build two dictionaries to store the word swaps we will do at the end. (Token text is immutable in spaCy.)
    # We can simultaneously build the second text's word-by-part-of-speech dict and its word swap dict
    text1_word_swaps, text2_word_swaps = {}, {}
    for token in doc2:
        if token.pos_ in parts_of_speech:
            doc2_words_keyed_by_pos[token.pos_].append((token.text, token.tag_))
            try:
                replacement_word, replacement_word_tag = doc1_words_keyed_by_pos[token.pos_].pop()
                replacement_word = reconcile_replacement_word(token.text_with_ws, token.tag_, replacement_word,
                                                              replacement_word_tag)
                text2_word_swaps[token.text_with_ws] = replacement_word
            except IndexError:  # There are no more words to substitute; the other text had more words of this p.o.s.
                pass
    for pos in parts_of_speech:
        random.shuffle(doc2_words_keyed_by_pos[pos])
    for token in doc1:
        if token.pos_ in parts_of_speech:
            try:
                replacement_word, replacement_word_tag = doc2_words_keyed_by_pos[token.pos_].pop()
                replacement_word = reconcile_replacement_word(token.text_with_ws, token.tag_, replacement_word,
                                                              replacement_word_tag)
                text1_word_swaps[token.text_with_ws] = replacement_word
            except IndexError:  # There are no more words to substitute; the other text had more words of this p.o.s.
                pass
    # Recompose the text from its whitespace-aware tokens, substituting words if needed.
    text1 = ''.join([text1_word_swaps.get(token.text_with_ws, token.text_with_ws) for token in doc1])
    text2 = ''.join([text2_word_swaps.get(token.text_with_ws, token.text_with_ws) for token in doc2])
    return text1, text2


def markov(input: input_type, ngram_size=1, num_output_sentences=5) -> List[str]:
    """Markov chain text generation from markovify library, supports custom n-gram length

    Keyword arguments:
    n-gram size: determines what n-gram model to use: x where x is order-x n-gram
    num_output_sentences: number of sentencess to output
    """
    if type(input) == list:
        list_of_texts = input
    elif type(input) == str:
        list_of_texts = [input]
    markov_models = []
    for text in list_of_texts:
        markov_models.append(markovify.Text(text, state_size=ngram_size))
    textgen = markovify.combine(markov_models)
    output_sentences = []
    while len(output_sentences) < num_output_sentences:
        sentence = textgen.make_sentence()
        if isinstance(sentence, str):
            output_sentences.append(sentence)
    return output_sentences


def cutup(input, min_cutout_words=3, max_cutout_words=7) -> List[str]:
    """Simulates William S. Burroughs' and Brion Gysin's cut-up technique by separating an input text into
    non-whitespace blocks of text and then randomly grouping those into cut-outs between the minimum and maximum
    length of words.

    Arguments:
        input (str) -- input string to be cut up
        min_cutout_words (int) -- minimum number of words in cut out chunk
        max_cutout_words -- maximum number of words in cutout chunk
    """
    if type(input) == list:
        list_of_texts = input
    elif type(input) == str:
        list_of_texts = [input]
    # We don't need tokenization for this since physically cutting up text out of books always cuts where whitespace
    # exists--it does not separate words from punctuation as punctuation does. (Also this way is faster.)
    cutouts = []
    for text in list_of_texts:
        word_list = text.split(" ")
        current_position, next_position = 0, 0
        while next_position < len(word_list):
            cutout_word_count = random.randint(min_cutout_words, max_cutout_words)
            next_position = current_position + cutout_word_count
            cutouts.append(" ".join(word_list[current_position:next_position]))
            current_position = next_position
    random.shuffle(cutouts)
    return cutouts
